﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
// type definitions for convenience
using HeuristicLab.Problems.DataAnalysis.Symbolic.Regression;
using CloneMapType = HeuristicLab.Core.ItemDictionary<HeuristicLab.Core.IItem, HeuristicLab.Core.IItem>;
using TraceMapType = HeuristicLab.Core.ItemDictionary<HeuristicLab.Core.IItem, HeuristicLab.Core.IItemList<HeuristicLab.Core.IItem>>;

namespace HeuristicLab.EvolutionaryTracking {
  /// <summary>
  /// An operator that gathers information on tree fragments that get passed from one generation to the next via genetic operators
  /// (Tracking of genetic variance and heredity)
  /// </summary>
  [Item("SymbolicExpressionTreeFragmentsAnalyzer", "An operator that provides statistics about crossover fragments")]
  [StorableClass]
  public abstract class SymbolicExpressionTreeFragmentsAnalyzer : SingleSuccessorOperator, IAnalyzer {
    #region Parameter names
    protected const string UpdateIntervalParameterName = "UpdateInterval";
    protected const string UpdateCounterParameterName = "UpdateCounter";
    protected const string ResultsParameterName = "Results";
    protected const string SecondaryTraceMapParameterName = "SecondaryTraceMap";
    protected const string SecondaryFragmentMapParameterName = "SecondaryFragmentMap";
    protected const string SecondaryCloneMapParameterName = "SecondaryCloneMap";
    protected const string GlobalTraceMapParameterName = "GlobalTraceMap";
    protected const string GlobalTraceMapParameterDescription = "Global map of trees and their parents.";
    protected const string GlobalCloneMapParameterName = "GlobalCloneMap";
    protected const string GlobalCloneMapParameterDescription = "Global map of trees and their clones";
    protected const string GlobalFragmentMapParameterName = "GlobalFragmentMap";
    protected const string GlobalFragmentMapParameterDescription = "Global map of trees and their respective fragments.";
    protected const string GenerationsParameterName = "Generations";
    protected const string SymbolicExpressionInterpreterParameterName = "SymbolicExpressionTreeInterpreter";
    protected const string SymbolicRegressionProblemDataParameterName = "ProblemData";
    #endregion

    // impact values calculator
    private readonly SymbolicRegressionSolutionImpactValuesCalculator calculator = new SymbolicRegressionSolutionImpactValuesCalculator();

    #region Parameter properties
    public ValueParameter<IntValue> UpdateIntervalParameter {
      get { return (ValueParameter<IntValue>)Parameters[UpdateIntervalParameterName]; }
    }
    public ValueParameter<IntValue> UpdateCounterParameter {
      get { return (ValueParameter<IntValue>)Parameters[UpdateCounterParameterName]; }
    }
    public LookupParameter<ResultCollection> ResultsParameter {
      get { return (LookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public LookupParameter<IntValue> GenerationsParameter {
      get { return (LookupParameter<IntValue>)Parameters[GenerationsParameterName]; }
    }
    // tracking structures
    public LookupParameter<TraceMapType> GlobalTraceMapParameter {
      get { return (LookupParameter<TraceMapType>)Parameters[GlobalTraceMapParameterName]; }
    }
    public LookupParameter<CloneMapType> GlobalCloneMapParameter {
      get { return (LookupParameter<CloneMapType>)Parameters[GlobalCloneMapParameterName]; }
    }
    public LookupParameter<CloneMapType> GlobalFragmentMapParameter {
      get { return (LookupParameter<CloneMapType>)Parameters[GlobalFragmentMapParameterName]; }
    }
    // auxiliary structures for tracking fragments and individuals from the previous generation
    // (this is needed because tracking is done in connection to the current generation, i.e., for 
    // counting "successful" offspring and fragments
    public LookupParameter<TraceMapType> SecondaryTraceMapParameter {
      get { return (LookupParameter<TraceMapType>)Parameters[SecondaryTraceMapParameterName]; }
    }
    public LookupParameter<CloneMapType> SecondaryFragmentMapParameter {
      get { return (LookupParameter<CloneMapType>)Parameters[SecondaryFragmentMapParameterName]; }
    }
    public LookupParameter<CloneMapType> SecondaryCloneMapParameter {
      get { return (LookupParameter<CloneMapType>)Parameters[SecondaryCloneMapParameterName]; }
    }
    // problem data, interpreter and evaluator
    public LookupParameter<SymbolicDataAnalysisExpressionTreeInterpreter> SymbolicExpressionInterpreterParameter {
      get { return (LookupParameter<SymbolicDataAnalysisExpressionTreeInterpreter>)Parameters[SymbolicExpressionInterpreterParameterName]; }
    }
    public LookupParameter<RegressionProblemData> SymbolicRegressionProblemDataParameter {
      get { return (LookupParameter<RegressionProblemData>)Parameters[SymbolicRegressionProblemDataParameterName]; }
    }
    #endregion

    #region Parameters
    public bool EnabledByDefault {
      get { return true; }
    }
    public IntValue UpdateInterval {
      get { return UpdateIntervalParameter.Value; }
    }
    public IntValue UpdateCounter {
      get { return UpdateCounterParameter.Value; }
    }
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    public CloneMapType GlobalCloneMap {
      get { return GlobalCloneMapParameter.ActualValue; }
    }
    public TraceMapType GlobalTraceMap {
      get { return GlobalTraceMapParameter.ActualValue; }
    }
    public TraceMapType SecondaryTraceMap {
      get { return SecondaryTraceMapParameter.ActualValue; }
      set { SecondaryTraceMapParameter.ActualValue = value; }
    }
    public CloneMapType SecondaryFragmentMap {
      get { return SecondaryFragmentMapParameter.ActualValue; }
      set { SecondaryFragmentMapParameter.ActualValue = value; }
    }
    public CloneMapType SecondaryCloneMap {
      get { return SecondaryCloneMapParameter.ActualValue; }
      set { SecondaryCloneMapParameter.ActualValue = value; }
    }
    public CloneMapType GlobalFragmentMap {
      get { return GlobalFragmentMapParameter.ActualValue; }
    }
    public IntValue Generations {
      get { return GenerationsParameter.ActualValue; }
    }
    public SymbolicDataAnalysisExpressionTreeInterpreter SymbolicExpressionInterpreter {
      get { return SymbolicExpressionInterpreterParameter.ActualValue; }
    }
    public RegressionProblemData SymbolicRegressionProblemData {
      get { return SymbolicRegressionProblemDataParameter.ActualValue; }
    }
    #endregion

    [StorableConstructor]
    protected SymbolicExpressionTreeFragmentsAnalyzer(bool deserializing) : base(deserializing) { }

    protected SymbolicExpressionTreeFragmentsAnalyzer(SymbolicExpressionTreeFragmentsAnalyzer original, Cloner cloner) : base(original, cloner) { }

    protected SymbolicExpressionTreeFragmentsAnalyzer()
      : base() {
      #region Add parameters
      // analyzer update counter and update interval
      Parameters.Add(new ValueParameter<IntValue>(UpdateIntervalParameterName, "The interval in which the tree length analysis should be applied.", new IntValue(1)));
      Parameters.Add(new ValueParameter<IntValue>(UpdateCounterParameterName, "The value which counts how many times the operator was called since the last update", new IntValue(0)));
      // tracking
      Parameters.Add(new LookupParameter<TraceMapType>(GlobalTraceMapParameterName, GlobalTraceMapParameterDescription));
      Parameters.Add(new LookupParameter<CloneMapType>(GlobalCloneMapParameterName, GlobalCloneMapParameterDescription));
      Parameters.Add(new LookupParameter<CloneMapType>(GlobalFragmentMapParameterName, GlobalFragmentMapParameterDescription));
      // secondary tracking 
      Parameters.Add(new LookupParameter<TraceMapType>(SecondaryTraceMapParameterName, GlobalTraceMapParameterDescription));
      Parameters.Add(new LookupParameter<CloneMapType>(SecondaryCloneMapParameterName, GlobalCloneMapParameterDescription));
      Parameters.Add(new LookupParameter<CloneMapType>(SecondaryFragmentMapParameterName, GlobalFragmentMapParameterDescription));
      // impact calculation
      Parameters.Add(new LookupParameter<SymbolicDataAnalysisExpressionTreeInterpreter>(SymbolicExpressionInterpreterParameterName, "Interpreter for symbolic expression trees"));
      Parameters.Add(new LookupParameter<RegressionProblemData>(SymbolicRegressionProblemDataParameterName, "The symbolic data analysis problem."));
      // fragment statistics parameters
      Parameters.Add(new ValueLookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
      Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName, "The number of generations so far."));
      #endregion

      UpdateCounterParameter.Hidden = true;
      UpdateIntervalParameter.Hidden = true;

    }
    #region After deserialization code
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      // check if all the parameters are present and accounted for 
      if (!Parameters.ContainsKey(UpdateIntervalParameterName)) {
        Parameters.Add(new ValueParameter<IntValue>(UpdateIntervalParameterName, "The interval in which the tree length analysis should be applied.", new IntValue(1)));
      }
      if (!Parameters.ContainsKey(UpdateCounterParameterName)) {
        Parameters.Add(new ValueParameter<IntValue>(UpdateCounterParameterName, "The value which counts how many times the operator was called since the last update", new IntValue(0)));
        UpdateCounterParameter.Hidden = true;
      }
    }
    #endregion
    #region IStatefulItem members
    public override void InitializeState() {
      base.InitializeState();
      UpdateCounter.Value = 0;
    }

    public override void ClearState() {
      base.ClearState();
      UpdateCounter.Value = 0;
    }
    #endregion

    /* A small description of the way this analyzer works
     * 
     * We keep 3 maps in the global scope: the GlobalTraceMap, the GlobalFragmentMap and the GlobalCloneMap that are updated on each step: selection, crossover, mutation
     * These maps provide information about the children, parents and transferred fragments as follows:
     * 
     * 1) GlobalCloneMap
     *    This data structure provides the basis for tracking. In the selection step, every selected individual gets cloned and transferred into the recombination pool.
     *    The GlobalCloneMap keeps track of clones by mapping them to the original individuals. This is an essential step for building the genealogy graph. Using this
     *    structure we can find out how many times each individual was selected. 
     *    The GlobalCloneMap consists of Key-Value pairs Clone->Original
     *
     * 2) GlobalTraceMap
     *    Keeps information in the form of key-value pairs Child->{Parents}, where {Parents} is a set consisting of 
     *      - one parent individual (in the case of mutation), by language abuse we say that the individual that mutation acted on is the parent,
     *        and the resulting individual is the child
     *      - two parent individuals (in the case of crossover), named Parent0 and Parent1. Parent0 is the individual that crossover acts on by replacing 
     *        one of its subtrees with a compatible subtree, randomly selected from Parent1
     *
     *    CROSSOVER
     *        Parent0     Parent1                                                                    {Parents}
     *           \           /                                                                           ^
     *            \       fragment (inserted from Parent1 into Parent0 to create the Child)              ^    [Mapping provided by the GlobalTraceMap]
     *             \       /                                                                             ^
     *              \     /                                                                              ^
     *               Child                                                                             Child
     *               
     *    MUTATION
     *        Parent0
     *           |
     *           |    [mutation acts on a random node/subtree in the 'parent' individual
     *           |
     *         Child
     *         
     *    Since crossover and mutation act on individuals in the recombination pool (which in fact are clones of individuals from the previous generation), the GlobalTraceMap
     *    is populated with values given by the mappings in the GlobalCloneMap, so that the original parent for each child is known.
     *    
     *  3) GlobalFragmentMap
     *     Keeps information in the form of Key-Value pairs about an individual and its respective fragment 
     *     (ie., the subtree received via crossover or created/altered by mutation)
     * */

    public override IOperation Apply() {
      // save the current generation so it can be compared to the following one, next time the analyzer is applied
      SecondaryTraceMap = SecondaryTraceMap ?? new TraceMapType();
      SecondaryCloneMap = SecondaryCloneMap ?? new CloneMapType();
      SecondaryFragmentMap = SecondaryFragmentMap ?? new CloneMapType();

      SecondaryTraceMap.Clear();
      if (GlobalTraceMap != null)
        foreach (var m in GlobalTraceMap)
          SecondaryTraceMap.Add(m.Key, m.Value);

      SecondaryCloneMap.Clear();
      if (GlobalCloneMap != null)
        foreach (var m in GlobalCloneMap)
          SecondaryCloneMap.Add(m.Key, m.Value);

      SecondaryFragmentMap.Clear();
      if (GlobalFragmentMap != null)
        foreach (var m in GlobalFragmentMap)
          SecondaryFragmentMap.Add(m.Key, m.Value);

      return base.Apply();
    }

    protected virtual void InitializeParameters() {
      #region Init parameters

      #endregion
    }

    private static int VisitationLength(ISymbolicExpressionTree tree) {
      return VisitationLength(tree.Root);
    }

    private static int VisitationLength(ISymbolicExpressionTreeNode node) {
      return node.IterateNodesBreadth().Sum(n => n.GetLength());
    }

  } //SymbolicExpressionTreeFragmentsAnalyzer
}
